CON

'' PropCOMM VT100_VGA W/PropDOS V2.3c By Jeff Ledger
''
'' This version of PropCOMM2 was written for use with the N8VEM board.
'' it is designed to allow the Propeller and Z80 to interact with each other.
''
'' The madness behind this program is two-fold:

'' 1) Provide the N8VEM guys a Keyboard/VGA VT100 terminal with SD storage.
'' 2) Pave the way for Propeller users to eventually use Wordstar for
''    the creation of .spin programs without connecting to a PC.
''
'' Xmodem Protocol Notes:    (Because someone will want to use it.)
'' ----------------------
'' The Xmodem Receive code was borrowed from Mike Cook's "sdxmodem.spin"
'' (I believe it was originally used in "xmodemVGA_BMP.spin", but I can't
'' tell where it originated, but Mike's version was very well commentted.
''
'' The Xmodem Send code was created based on
'' "Understanding The X-Modem File Transfer Protocol" by: Em Decay
'' I tried to follow Mike's lead and comment it well.
''
'' Protocol is provided AS IS.  All I can tell you is that it works
'' with XMODEM38 Version 5.0 under CP/M.  Which is what I indended. :-)
''
'' VT100 Emulation Notes:
'' ----------------------
'' The VT100 terminal code supports all of the BASIC cursor positioning
'' codes as well as home, cleardown, inverse, and bold.  It is intended
'' to provide the "commonly used" codes required for products like Wordstar.

   _clkmode  = xtal1 + pll16x
   _xinfreq  = 5_000_000

   termcolor = LIGHTAMBER             ' Pick from terminal colors below.
   keyboard  = 26                     ' Keyboard input
   video     = 16                     ' Video output

   serXmit   = 9                      ' Serial Transmit
   serRecv   = 8                      ' Serial Receive
   baud      = 19200                  ' Baud Rate

   spiDO     = 0                      ' SD I/O
   spiClk    = 1                      ' SD I/O
   spiDI     = 2                      ' SD I/O
   spiCS     = 3                      ' SD I/O


CON

   ''Terminal Colors
   DIMCYAN           = $29
   DIMGREEN          = $27
   DIMORANGE         = $92
   DIMPINK           = $95
   DIMRED            = $C1
   PURPLE            = $99
   CLASSICAMBER      = $A2
   CLASSICAMBERDARK  = $E2
   LIGHTAMBER        = $A5
   WHITE             = $FF
   HOTPINK           = $C9
   PINK              = $D9
   RED               = $C5
   CYAN              = $7F
   DARKBLUE          = $5F
   CLASSICGREEN      = $75

   CR                = $0D
   LF                = $0A
   TAB               = $09
   BKS               = $08

   bspKey            = $C8
   escape            = $CB
   backspace         = $C8
   fReturn           = 13
   fLinefeed         = 10
   fEof              = -1

   '' Xmodem Vars

   SOH               = 1               ' Packet Start
   STX               = 2
   EOT               = 4               ' End of Transmission
   ACK               = 6               ' Positive Acknowledgement
   NAK               = 21              ' Negative Acknowledgement
   CAN               = 24              ' Cancel
   PAD               = 26              ' Padding at the end of the file
   RX_Buffer         = 20


   debugvideo        = 12
   debugyellow       = $1E  'debug VIDEO SENT TO TV --
   debugwhite        = $AF  ''SHOULD NOT HURT ANYTHING
OBJ
  ''debug  : "aigeneric_driver"         ' debug Screen (TV)
  text   : "VGA_1024"                  ' VGA Terminal Driver
  key    : "comboKeyboard"             ' Keyboard Driver
  ser    : "FullDuplexSerial256"       ' Full Duplex Serial Controller
  fsrw   : "fsrwFemto"                 ' Modified SD routines

VAR
     byte tbuf[20]                     ' SD Vars
     long ioControl[2]

     byte command[32]                  ' PropDOS command line Vars
     byte cline[42]
     byte strbuff[5]

     byte cursor                       ' Cursor keys setting Num/WS

     BYTE CRC                          ' XMODEM Protocol
     BYTE PACKET                       ' XMODEM Protocol
     BYTE XRECV                        ' XMODEM Protocol
     BYTE MODE                         ' XMODEM Protocol
     BYTE RETRY                        ' XMODEM Protocol
     BYTE XBUF[128]                    ' XMODEM Protocol
     BYTE pdata[1028]                  ' XMODEM Protocol



PUB start | i, a, b, TimeCnt

  ''debugstart(debugvideo)                       ' Start debug driver
  ''debugstr(string("debug READY",13,13))

  key.start(keyboard)                           ' Start keyboard driver
  text.start(video)
  fsrw.start(@ioControl)                        ' Start I2C/SPI driver                             ' Start terminal screen

  ser.start(serRecv,serXmit,0,baud)             ' Control link via prop-plug

  'cursor := 0  ' 0=Cursor keys match number keys
  cursor := 1  ' 1=Cursor keys match Wordstar Movement keys

  '' Check for the required keyboard & SD card, gracefull abort if not found.
  if \fsrw.mount(spiDO,spiClk,spiDI,spiCS) < 0
     text.str(string(13,"No SD card present.",CR))
     text.str(string("SD card expected on: "))
     text.dec(spiDO)
     text.str(string(", "))
     text.dec(spiClk)
     text.str(string(", "))
     text.dec(spiDI)
     text.str(string(", "))
     text.dec(spiCS)
     text.str(string(13,CR,"System halted.",CR))
     abort



  titlescreen
  ser.tx(13)  ''A quick wake up to the remote device.
  terminal

PRI terminal | filebyte,filecount, x,local,record,remote,vt100,byte2,byte3,byte1,byte4,byte5,byte6,byte7,loop,var1,col,row

  vt100 := 0
  bytefill(@cline,0,14) ' Clear commandline buffer

  repeat
      local := key.key
      if local > 0

         if local > 608         '' Assume CTRL key is being held
            local:=local-608    '' Redefine as CTRL+char

         if local == 203        '' Check for [ESC] Key
            local := 27         '' Change to standard ASCII value

         if local == 208        '' Check for [F1] and clearscreen
            local := -1         '' Redisplay opening splash page.
            titlescreen
            ser.tx(13)          ''A quick wake up to the remote device.

         if local == 209        '' Check for [F2]
            local := -1         '' and send DIR command
            dirremote


         if local == 217 ' F10 Switch to PropDOS type prompt.
            text.out(13)
            propdos

         if local == 219 and cursor == 0 ' F12 Switch Cursor mode to WS
            cursor:=1
            local := -1
            text.str(string("Cursor keys in Wordstar Mode."))
            waitcnt(100_000_000 + cnt)
            x:=0
            repeat x from 0 to 56
              text.out($C8)
              x++

         if local == 219 and cursor == 1 ' F12 Switch Cursor mode to Numbers
            cursor:=0
            local := -1
            text.str(string("Cursor keys in Numbers Mode."))
            waitcnt(100_000_000 + cnt)
            x:=0
            repeat x from 0 to 54
              text.out($C8)
              x++

         if local == 200 or local == 201 or local == 8
            'local := 08
            'text.str(string($C8," "))
            local:=08

         '' Number Pad compatibility for the cursor keys
         if local == 194 and cursor == 0'' If cursor up
            local := 56  '' Send an 8

         if local == 195 and cursor == 0'' If cursor down
            local := 50  '' Send a 2

         if local == 192 and cursor == 0'' If cursor left
            local := 52  '' Send a 4

         if local == 193 and cursor == 0'' If cursor right
            local := 54  '' Send a 6


         '' Wordstar compatibility for the cursor keys
         if local == 194 and cursor == 1'' If cursor up
            local := 5   '' Send a ^E

         if local == 195 and cursor == 1'' If cursor down
            local := 24  '' Send a ^X

         if local == 192 and cursor == 1'' If cursor left
            local := 19  '' Send a ^S

         if local == 193 and cursor == 1'' If cursor right
            local := 4   '' Send a ^D


         '' 197 END
         '' 196 HOME

         ser.tx(local)
         local := -1

      '## Deal with remote data if present ##


    remote := ser.rxcheck
     if (remote > 0)

      if remote == 15
         remote:=0

      if remote > 0         '' Used for debugging.  Can be removed.
        ''debugout(remote)   '' Display everthing on the TV terminal
       '''debugdec(local)


      if remote == 27 or remote == 124 'vt100 ESC code is being sent
         vt100:=1
         byte1:=0
         byte2:=0
         byte3:=0
         byte4:=0
         byte5:=0
         byte6:=0
         byte7:=0
         remote:=0 'Don't display the ESC code
         ''debugcolor(debugyellow)
         ''debugout(92)
         ''debugcolor(debugwhite)

      if remote == 91 and vt100 == 1 'look for open bracket [
         vt100:=2 'start recording code

      if remote == 62 and vt100 == 1 or remote == 60 and vt100 == 1'look for < & >
         vt100:=0 ' not sure why this is coming up, can't find in spec.

      if vt100==2 ''Check checking for VT100 emulation codes


         if remote > 10
           byte7:=byte6
           byte6:=byte5 '' My VTCode Mini Buffer
           byte5:=byte4
           byte4:=byte3
           byte3:=byte2 ''Record the last 7 bytes
           byte2:=byte1
           byte1:=remote



         if remote == 109         ' Look for lowercase m
            if byte2 == 91        ' if [m turn off to normal set
               text.inv(0)
            if byte2 == 49        ' is it ESC[1m BOLD
               text.inv(1)
            if byte2 == 55        ' is it ESC[7m REVERSE
               text.inv(1)
            if byte2 == 48        ' 0 is back to normal
               text.inv(0)
            vt100:=0



         '' ***>> This section is specific to PropDOS Batch Commands <<***
         '' ***>> It uses a NON-STANDARD VT100 type command to invoke <***
         '' ***>> ESC[RX Xmodem Receive file from remote
         '' ***>> ESC[SX Xmodem Send file to remote

         if remote == 88         ' Look for UPPERCASE X
            if byte2 == 82       ' Look for UPPERCASE R 'Initiate remote Send
               '' Start file fetch loop
               bytefill(@cline,0,14)              '' Clear commandline buffer
               filecount:=0
               filebyte:=0
               repeat until filebyte == 13        '' ESC[RX Understood.
                  filebyte:= ser.rxcheck          '' Look for filename.

                  if filebyte > 34
                     cline[filecount]:=filebyte
                     filecount++
               waitcnt(clkfreq/100 + cnt)
               XUPLOAD                            '' Trigger Xmodem Upload

               vt100:=0

            if byte2 == 83        ' Look for UPPERCASE S 'Initiate remote Receive
               '' Start file fetch loop
               bytefill(@cline,0,14)              '' Clear commandline buffer
               filecount:=0
               filebyte:=0
               repeat until filebyte == 13        '' ESC[SX Understood.
                  filebyte:= ser.rxcheck          '' Look for filename.

                  if filebyte > 34
                     cline[filecount]:=filebyte
                     filecount++
               waitcnt(clkfreq/100 + cnt)
               XDOWNLOAD                          '' Trigger Xmodem Download

               vt100:=0


         '' **************************************************************




         '' Remove some unsupported codes from being shown.

         if remote == 104 'look for lowercase h
            vt100:=0

         if remote == 61  'lool for =
            vt100:=0

         if remote == 114 'look for lowercase r
            vt100:=0

         if remote == 108 'look for lowercase l
            vt100:=0

         if remote == 62  'look for >
            vt100:=0

         if remote == 77  'Look for M (SCROLL UP ONE LINE)
            text.scrollM
            vt100:=0

         if remote == 76 'Look for L (SCROLL DOWN ONE LINE)
            text.scrollD
            vt100:=0


         if remote == 72 or remote == 102'' HOME CURSOR (uppercase H or lowercase f)
            if byte2==91 or byte2==59   ''look for [H or ;f
               text.home

            '' Check for X & Y with [H or ;f   -   Esc[Line;ColumnH

            else
              if byte4 == 59 '' Assume that col is greater than 9
                byte3:=byte3-48    'Grab 10's
                byte2:=byte2-48    'Grab 1's
                byte3:=byte3*10    'Multiply 10's
                byte3:=byte3+byte2 'Add 1's
                col:=byte3         'Set cols

                if byte7 == 91''Assume row number is greater than 9
                   byte6:=byte6-48 'Grab 10's
                   byte5:=byte5-48 'Grab 1's
                   byte6:=byte6*10 'Multiply 10's
                   byte6:=byte6+byte5 'Add 1's
                   row:=byte6

                if byte6 == 91'Assume row number is less than 10
                   byte5:=byte5-48 'Grab 1's
                   row:=byte5

              if byte3 == 59 '' Assume that col is less an 10
                byte2:=byte2-48    'Grab 1's
                col:=byte2         'set cols

                if byte6 == 91''Assume row number is greater than 9
                   byte5:=byte5-48 'Grab 10's
                   byte4:=byte4-48 'Grab 1's
                   byte5:=byte5*10 'Multiply 10's
                   byte5:=byte5+byte4 'Add 1's
                   row:=byte5
                if byte5 == 91 'Assume that col is greater than 10
                   byte4:=byte4-48 'Grab 1's
                   row:=byte4

              col:=col-1

              if col == -40    ' Patches a bug I havn't found.  *yet*
                 col := 58     ' A Microsoft approach to the problem. :)
              if row == -459
                 row := 1
              if row == -449
                 row := 2      ' Appears to be an issue with reading
              if row == -439   ' single digit rows.
                 row := 3
              if row == -429   ' This patch checks for the bug and replaces
                 row := 4      ' the faulty calculation.
              if row == -419
                 row := 5      ' Add to list to find the source of bug later.
              if row == -409
                 row := 6
              if row == -399
                 row := 7
              if row == -389
                 row := 8
              if row == -379
                 row := 9


              text.cursloc(col,row)

            vt100:=0

         if remote == 74    '' CLEAR SCREEN
            if byte2==91    '' look for [J  '' clear screen from cursor to 25
               text.clsfromcursordown
            vt100:=0
            if byte2==50    '' look for [2J '' clear screen
               text.cls



         if remote == 66    '' CURSOR DOWN    Esc[ValueB
            if byte4 == 91 '' Assume number over 10
              byte3:=byte3-48
              byte2:=byte2-48
              byte3:=byte3*10
              byte3:=byte3+byte2
              var1:=byte3
            if byte3 == 91 '' Assume number is less 10
              byte2:=byte2-48
              var1:=byte2

              'var1:=var1-1 '' Seems to move too far with B2 ''May remove



            loop:=0
            repeat until loop == var1
               loop++
               text.out($C3)

            vt100:=0


         if remote == 65    '' CURSOR UP   Esc[ValueA
            if byte4 == 91 '' Assume number over 10
              byte3:=byte3-48
              byte2:=byte2-48
              byte3:=byte3*10
              byte3:=byte3+byte2
              var1:=byte3
            if byte3 == 91 '' Assume number is less 10
              byte2:=byte2-48
              var1:=byte2

            loop:=0
            repeat until loop == var1
               text.out($C2)
               loop++
            vt100:=0


         if remote == 67    '' CURSOR RIGHT   Esc[ValueC
            if byte4 == 91 '' Assume number over 10
              byte3:=byte3-48
              byte2:=byte2-48
              byte3:=byte3*10
              byte3:=byte3+byte2
              var1:=byte3
            if byte3 == 91 '' Assume number is less 10
              byte2:=byte2-48
              var1:=byte2

            loop:=0
            repeat until loop == var1
               text.out($C1)
               loop++
            vt100:=0



         if remote == 75   '' Clear line  Esc[K
           if byte2 == 91 or byte2 == 48 '' Look for [  or 0
             text.clearlinefromcursor


         remote:=0 '' hide all codes from the VGA output.

      if remote == 8
         text.out(backspace)
         remote := 0

      if record == 13 and remote == 13  ''remove LF
         remote :=0

      if remote > 11
           text.out(remote)

      record:=remote ''record last byte

PUB propdos  | counter, newcount,r,rr,tt

 repeat
   bytefill(@command,0,32) 'Clear the command buffer
   bytefill(@cline,0,14) ' Clear commandline buffer

   text.str(string("P>"))

   input(@command,32)

   if textComp(@command,string("spin"),4) == true
     counter := 5 ' Starting point of command line (after space)
     newcount := 0
     repeat until newcount > 14
       cline[newcount] := command[counter]
       counter++
       newcount++
    fsrw.popen(@cline,"r")
    fsrw.bootSDCard

   elseif textComp(@command,string("xmodem r "),9) == true    'Xmodem download from CP/M
     counter := 9 ' Starting point of command line (after space)
     newcount := 0
     ser.str(string("b:xmodem s "))    'assumes xmodem is on B:
     repeat until newcount > 14
       cline[newcount] := command[counter]
       ser.tx(command[counter])
       counter++
       newcount++
     ser.tx(13)
     xdownload

   elseif textComp(@command,string("xmodem s "),9) == true    'Xmodem upload to CP/M
     counter := 9 ' Starting point of command line (after space)
     newcount := 0
     ser.str(string("era "))          'erase file if it already exists
     repeat until newcount > 14       'to prevent xmodem from jamming.
       cline[newcount] := command[counter]
       ser.tx(command[counter])
       counter++
       newcount++
     waitcnt(clkfreq/100 + cnt)
     ser.tx(13)
     BATCHCOMMAND                      'wait for a > prompt.
     counter := 9 ' Starting point of command line (after space)
     newcount := 0
     ser.str(string("b:xmodem r "))    'assumes xmodem is on B:
     repeat until newcount > 14
       cline[newcount] := command[counter]
       ser.tx(command[counter])
       counter++
       newcount++
     waitcnt(clkfreq/100 + cnt)
     ser.tx(13)
     xupload


   elseif textComp(@command,string("dir"),3) == true or textComp(@command,string("$"),1) == true
      cpmdir

   elseif textComp(@command,string("reboot"),6) == true
     reboot

   elseif textComp(@command,string("a:"),2) == true
     ser.tx(13)
     terminal

   elseif textComp(@command,string("type"),4) == true
     counter := 5 ' Starting point of command line (after space)
     newcount := 0
     repeat until newcount > 14
       cline[newcount] := command[counter]
       counter++
       newcount++
     r := fsrw.popen(@cline,"r")
      repeat
        r := fsrw.pgetc
         if r < 0
           text.out(13)
           quit
        if r > 0
          text.out(r)

   elseif textComp(@command,string("supersub"),8) == true
     text.str(string(13,"Prodos SuperSUB V1.0",13,13))
     counter := 9 ' Starting point of command line (after space)
     newcount := 0
     repeat until newcount > 14
       cline[newcount] := command[counter]
       counter++
       newcount++
     r := fsrw.popen(@cline,"r")
      repeat
        r := fsrw.pgetc
         if r < 0
           text.out(13)
           quit
         if r > 12
           ser.tx(r)
           text.out(r)
         if r==13
           BATCHCOMMAND

   elseif textComp(@command,string("del"),3) or textComp(@command,string("era"),3) == true
     counter := 4 ' Starting point of command line (after space)
     newcount := 0
     repeat until newcount > 14
       cline[newcount] := command[counter]
       counter++
       newcount++
     counter := 0
     r := fsrw.popen(@cline,"d")

   elseif textComp(@command,string("help"),3) or textComp(@command,string("?"),1) == true
     text.str(string(13,"-:- PropCOMM2_VGA w/PropDOS for Z80 Interaction -:-",13,13))
     text.str(string(13,"PROPDOS COMMANDS:",13))
     text.str(string("dir        =  Show directory of Propeller SD",13))
     text.str(string("del or era =  Erase file (usage: era <filename>)",13))
     text.str(string("type       =  Diaplay text file (usage: type <filename>)",13))
     text.str(string("xmodem r   =  Download from CP/M (usage: xmodem r <filename>)",13))
     text.str(string("xmodem s   =  Upload to CP/M (usage: xmodem s <filename>)",13))
     text.str(string("spin       =  Launch spin binary (usage: spin <filename>)",13))
     text.str(string("supersub   =  Batch commands to A: (usage: supersub <filename>)",13))
     text.str(string("reboot     =  Reboot Propeller / Will launch EEPROM file.",13,13))
     text.str(string(".COM files can be executed from the P> prompt if they",13))
     text.str(string("exist on the A> disk.  Otherwise you'll get an file error",13))
     text.str(string("Use 'xmodem s' to send them to A> before executing.",13))
     text.out(13)


   elseif command[1] <> 0    '' AUTO-EXECUTE A .COM FILE IF IT EXISTS ON A>
      counter := 0
      repeat until counter > 13
         if command[counter] > 0
            '' Convert Lowercase to UPPERCASE
            if command[counter] > 96 and command[counter] < 123
               command[counter] := command[counter]-=32
            cline[counter] := command[counter]
         counter++

      counter:=strsize(@cline)
      counter--
      rr:=strsize(@cline)
      fsrw.opendir
      repeat while fsrw.nextfile(@tbuf) == 0
        tt:=(StrRight(@tbuf,3))
        if textcomp(tt,string("COM"),3) == true
           if textComp(@cline,@tbuf,rr) == true
              'fsrw.popen(@cline,"r")
              'fsrw.bootSDCard
               ser.tx(13)
               waitcnt(clkfreq/2 + cnt)
               ser.str(@cline)
               waitcnt(clkfreq/100 + cnt)
               ser.tx(13)
               terminal



PRI cpmdir | r, sta, bytes,tt, dirleng
   text.out(13)
   dirleng:=0
   fsrw.opendir
   text.str(string("P: "))
   repeat while fsrw.nextfile(@tbuf) == 0
      if dirleng > 60
        text.out(13)
        text.str(string("P: "))
        dirleng:=0

     text.str(@tbuf)
     repeat 13 - strsize(@tbuf)
        text.str(string(" "))

     if dirleng < 40
        text.str(string(": "))
     dirleng := dirleng + 13

   text.out(13)

PUB dirremote
    ser.tx(13)
    waitcnt(clkfreq/2 + cnt)
    ser.str(string("DIRX"))
    waitcnt(clkfreq/100 + cnt)
    ser.tx(13)

PUB input(data, size) | keystroke,len,junk
   keystroke:=0
   len:=0
   text.out(02)
   repeat
     keystroke := key.key        '' Fetch key from keyboard if exists
     junk := ser.rxcheck         '' Fetch left over garbage from remote & disregard.

      if keystroke == 208        '' Check for [F1] and clearscreen
         keystroke := -1         '' Redisplay opening splash page.
         titlescreen
         text.out(13)
         quit

      if keystroke == 209
         cpmdir
         quit

      if keystroke == 217        ' F10 Switch to PropDOS type prompt.
         ser.tx(13)
         terminal

      if keystroke == $C8        ' Check for Backspace Key
        if len > 0
         byte[data--] := 0
         size++
         len--
         text.out(backspace)
         keystroke := 0

      if keystroke == $C8         'Clean up backspace if all is cleared
         keystroke:= 0

      if keystroke == 13          'Exit when Carriage Return is pressed
         text.out(13)
         quit

      if keystroke > 10
        if size > 1
          text.out(keystroke)


          ''text.dec(keystroke) 'for troubleshooting
          if keystroke <> 13
            byte[data++] := keystroke
            size--
            len++


PRI textComp(str1,str2,len)
   repeat len
      if byte[str1++] <> byte[str2++]
         return false
   return true

PUB StrRight(str,len) | i,l,z,j

  l:=strSize(str)
  z:=0
  repeat i from l-len to l
    strbuff[z] := byte[@tbuf][i]
    z++
  return @strbuff


PUB XDOWNLOAD | TCNT , temp, blockbar

  PACKET   := 1                      ' Transfer not yet started, Set Packet Number to first
  RETRY    := 0                      ' Clear Retry count
  MODE     := 0                      ' Set Mode to zero
  BLOCKBAR := 0

  REPEAT 300                         ' 40 seconds max time to wait for sender to send the file
    ser.tx(NAK)                      ' send NAK to sender to inform them we are ready to receive file
    XRECV := ser.rxtime (500)        ' wait here for character or time-out

    IF (XRECV == SOH)                ' If character received is packet start
      MODE := 1                      ' Set file transfer mode to state 1
      QUIT                           ' exit this repeat loop

  text.str(string(13,"Download Started:",13))
  fsrw.popen(@cline,"w")             ' open file for write, if exists, then over write the file

  REPEAT WHILE MODE == 1             ' While in file transfer state
    CRC := 0                         ' clear CRC for later calculation
    XRECV := ser.rx                  ' get packet number

    IF (XRECV < PACKET - 1) OR (XRECV > PACKET)  ' if packet number is in error
      MODE := 0                      ' Set to state zero
      QUIT                           ' exit this repeat loop
    ELSE
      PACKET := XRECV                ' next Packet Number

    XRECV := ser.rx                  ' 1's complement Packet Number
    TCNT := 0                        ' set for start of receive buffer
    REPEAT 128                       ' Get 128 Bytes from Sender
      XBUF[TCNT] := ser.rx           ' get character from sender in put in file transfer buffer
      CRC += XBUF[TCNT++]            ' update CRC and incriment file transfer buffer

    BLOCKBAR++
    text.str(string("="))            ' Do something to humor the user.
    if BLOCKBAR == 50                ' People love to be entertained.
      BLOCKBAR:=0
      text.out(13)

    XRECV := ser.rx                  ' get CRC from sender
    IF (XRECV <> CRC)                ' If CRC fail
      IF (RETRY < 10)                ' and retrys are less then 10
        ser.tx(NAK)                  ' Tell sender we got a bad packet
        RETRY++
        text.out(120)                ' Display a nasty X if error
        REPEAT                       ' Check for New Packet
          XRECV := ser.rx            ' get character from sender
          IF (XRECV == SOH)          ' check character for packet start
            QUIT                     ' exit this repeat loop

      ELSE                           ' Too many retrys
          MODE := 0                  ' set tranfer mode to fail
          QUIT                       ' exit this repeat loop

    ELSE                             ' Packet looks OK, lets wirte it to the SD Card
      ser.tx(ACK)                    ' Send ACK to sender
      fsrw.pwrite(@XBUF, 128)        ' first packet twice???, flag is used to filter and discard
      PACKET++                       ' the first xmodem packet.

      REPEAT                         ' Check were we are in the file transfer process
        XRECV := ser.rx
        IF (XRECV == EOT)            ' End of transmission ?
          ser.tx(ACK)                ' send ACK
          waitcnt(clkfreq/10 + cnt)  ' wait 100mS
          ser.tx (ACK)               ' send ACK again
          MODE := 2                  ' Set mode to completed successfully
          QUIT                       ' exit this repeat loop

        IF (XRECV == CAN)            ' Sender Cancelled transmission?
          MODE := 0                  ' Transfer aborted or incomplete
          QUIT                       ' exit this repeat loop

        IF (XRECV == SOH)            ' Next packet in the file
          QUIT

  'ser.str(@Newline)                 ' Send CR & LF to sender

  IF (MODE == 2)                     ' Notify sender of status of file transfer
    text.str (string(13,13,"Download Successful",13))                                      '   Upload was a success
  ELSE
    text.str (string(13,13,"Download Failed",13))                                      '   Failure in file transfer

  fsrw.pclose                        ' Close data file & exit
  ser.tx(13)
  ser.tx(13)

PUB XUPLOAD | i,d,blockbar

    '' BASED ON: Understanding The X-Modem File Transfer Protocol
    ''           by: Em Decay
    ''
    '' This works for my project, but YMMV.  Use at own risk. :)

  PACKET   := 0                      ' Transfer not yet started, Set Packet Number to first
  RETRY    := 0                      ' Clear Retry count
  MODE     := 0
  BLOCKBAR := 0

  REPEAT 300                         ' 40 seconds max time to wait for sender to send the file
    XRECV := ser.rx                  ' wait here for character or time-out
    'TEXT.OUT(XRECV)
     IF (XRECV == NAK)               ' If NAK is received is packet start
       MODE := 1                     ' Set file transfer mode to state 1
       QUIT


  fsrw.popen(@cline,"r")             ' Open local file for upload

  text.str(string(13,"Upload Started: ",13))

  REPEAT WHILE MODE == 1             ' While in file transfer state

    PACKET++                         ' Increment the PACKET

    IF PACKET > 255                  ' If Packet is higher than 255
      REPEAT UNTIL PACKET < 256      ' Recalculate PACKET number
         PACKET:=PACKET-256          ' Could be wrong, but works anyway. :

    CRC := 0                         ' clear CRC for later calculation
    ser.tx(SOH)                      ' SEND SOH BYTE (Step 1) NCGbyte
    ser.tx(PACKET)                   ' SEND PACKET # (Step 2)
    ser.tx(255-PACKET)               ' SEND ONES     (Step 3)

    repeat i from 0 to 127
      d:=fsrw.pgetc                  ' FETCH 128 BYTES FROM FILE

      if d==-1                       ' Check for end of file
        d:=EOT                       ' Send EOT
        MODE:=2                      ' Set Mode to Success & Bug out!

      pdata[i]:=d                    ' Position Byte in Packet
      CRC:=CRC+d                     ' Add the byte to CRC TOTAL

      ser.tx(pdata[i])               ' SEND THE PACKET (Step 4)

    repeat until CRC < 256           ' Calculate the checksum
      CRC:=CRC-256

    ser.tx(CRC)                      ' SEND THE CHECKSUM (Step 5)

    BLOCKBAR++
    text.str(string("="))            ' Do something to humor the user.

    if BLOCKBAR == 50                ' People love to be entertained.
      BLOCKBAR:=0
      text.out(13)

    repeat
      i:=ser.rx                      ' WAIT FOR ACK or NAK

      if i == 43                     ' Look for + sign, meaning some sort of error
        MODE:=0                      ' Switch to mode 0, and fail the upload
        quit                         ' These 3 lines are specific to N8VEM Xmodem.

      if i==ACK
        'text.out(65)                ' display A on ACK {remarked}
        RETRY:=0                     ' Reset the retry account
        quit
      if i==NAK
        'retransmit packet
        text.out(78)                 ' display N on NAK
        repeat i from 0 to 127
          ser.tx(pdata[i])           ' Send packet again
          ser.tx(CRC)                ' Send checksum again
        RETRY++                      ' Increment the retry
      if RETRY == 6                  ' Look for six concecutive fails.
        MODE:=0                      ' Set mode to indicate a fail
        quit                         ' Failed.

  ser.tx(EOT)                        ' Assume the ACK was recieved.
  'ser.tx(13)                        ' poor method, but it works.

  fsrw.pclose                        ' Close data file & exit

  IF (MODE == 2)                     ' Notify sender of status of file transfer
    text.str (string(13,13,"Upload Successful.",13))                                      '   Upload was a success
  ELSE
    text.str (string(13,13,"Upload Failed.",13))
  text.out(13)

PUB BATCHCOMMAND | BATCH, BATX
    '' This routine waits for the ">" to return from the remote
    '' system before returning.

    BATCH := 0
    BATX  := 0

    repeat
       BATCH:=SER.RX
         'text.out(BATCH)
       BATX++
       if BATCH==62                  '' look for a > indicating the remote is ready
         quit                        '' for the next command.
       if BATX==100                  '' If the loop runs too long,
         quit                        '' allow it to die gracefully.

PUB titlescreen

  text.cls
  text.color(termcolor)
  text.out(13)
  text.inv(1)
  text.str(string("  PROPCOMM VT100 W/PROPDOS SHELL   -:-   VERSION 2.3d  -:-    BY: JEFF LEDGER  "))
  text.inv(0)
  text.out(13)
  text.str(string(13,"["))
  text.inv(1)
  text.str(string("F1"))
  text.inv(0)
  text.str(string("] Clear screen.          -:- ["))
  text.inv(1)
  text.str(string("F2"))
  text.inv(0)
  text.str(string("] DIR Macro",13))


  text.str(string("["))
  text.inv(1)
  text.str(string("F10"))
  text.inv(0)
  text.str(string("] toggle terminal/shell -:- ["))
  text.inv(1)
  text.str(string("F12"))
  text.inv(0)
  text.str(string("] toggle cursorkeys (Num/WS mode)",13,13))
  if cursor == 0
    text.str(string("Cursor keys in: Numbers Mode.",13))
  ELSE
    text.str(string("Cursor keys in: Wordstar Mode.",13))

  text.str(string("Terminal Ready: "))
  text.dec(baud)
  text.str(string(",N,8,1",13))




